# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.

"""FMO Formatting Utilities."""

from __future__ import annotations

from typing import NoReturn

import typer


def secho_error_and_exit(text: str, color: str = typer.colors.RED) -> NoReturn:
    """Print error and exit."""
    typer.secho(text, err=True, fg=color)
    raise typer.Exit(1)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.
# pylint: disable=line-too-long, too-many-arguments, no-name-in-module
"""Frienli-Model-Optimizer (FMO) CLI."""

from __future__ import annotations

import os
from enum import Enum
from typing import Optional

import typer

from fmo.utils.dataset import get_calib_dataloader, get_tokenizer, safe_load_datasets
from fmo.utils.format import secho_error_and_exit
from fmo.utils.version import get_installed_version

app = typer.Typer(
    help="Friendli Model Optimizer 🚀",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
    add_completion=True,
    pretty_exceptions_enable=False,
)


class QuantMode(str, Enum):
    """Quant Mode."""

    INT8 = "int8"
    FP8 = "fp8"


@app.command()
def version():
    """Check the installed package version."""
    installed_version = get_installed_version()
    typer.echo(installed_version)


@app.command()
def quantize(
    model_name_or_path: str = typer.Option(
        ...,
        "--model-name-or-path",
        "-m",
        help="Hugging Face pretrained model name or directory path of the saved model checkpoint.",
    ),
    output_dir: str = typer.Option(
        ...,
        "--output-dir",
        "-o",
        help=(
            "Directory path to save the quantized checkpoint and related configurations "
            "files."
        ),
    ),
    mode: QuantMode = typer.Option(
        ...,
        "--mode",
        help=("Qantization techniques to apply. You can use `fp8`, and `int8`."),
    ),
    pedantic_level: int = typer.Option(
        1,
        "--pedantic-level",
        help=(
            "Higher pedantic level ensure a more accurate representation of the model,"
            "but increase the quantization processing time. Lower levels allow for faster"
            "quantization, but may sacrifice some model accuracy. Defaults to 1."
        ),
    ),
    device: Optional[str] = typer.Option(
        "cuda:0",
        "--device",
        help=(
            "The device which is used for quantization process. Currently we use only one GPU."
        ),
    ),
    offload: Optional[bool] = typer.Option(
        False,
        "--offload",
        help=(
            "When enabled, significantly reduces GPU memory usage by offloading model layers onto CPU RAM. Defaults to False."
        ),
    ),
    seed: Optional[int] = typer.Option(
        42, "--seed", help=("Seed for dataset sampling.")
    ),
    dataset_name_or_path: str = typer.Option(
        "abisee/cnn_dailymail:3.0.0",
        "--dataset-name-or-path",
        help=(
            "Huggingface dataset name or directory path for gathering sample activations."
        ),
    ),
    dataset_split_name: str = typer.Option(
        "test",
        "--dataset-split-name",
        help=("Huggingface dataset split name for gathering sample activations."),
    ),
    dataset_target_column_name: str = typer.Option(
        "article",
        "--dataset-target-column-name",
        help=("Huggingface dataset column name for gathering sample activations."),
    ),
    dataset_num_samples: int = typer.Option(
        512,
        "--dataset-num-samples",
        help=("The number of samples for gathering sample activations."),
    ),
    dataset_max_length: int = typer.Option(
        512,
        "--dataset-max-length",
        help=(
            "The maximum legth of sample in the dataset for gathering sample activations."
        ),
    ),
    dataset_batch_size: int = typer.Option(
        32,
        "--dataset-batch-size",
        help=("The batch size of the dataset for gathering sample activations."),
    ),
    cache_dir: Optional[str] = typer.Option(
        None, "--cache-dir", help="Directory path of the cached model checkpoint."
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", help="Only check quantization avaliability."
    ),
):
    """Quantize huggingface's model."""
    # pylint: disable=too-many-locals, import-outside-toplevel
    from fmo_core import quantize  # type: ignore

    # pylint: enable=import-outside-toplevel

    if not os.path.isdir(output_dir):
        if os.path.exists(output_dir):
            secho_error_and_exit(f"'{output_dir}' exists, but it is not a directory.")
        os.mkdir(output_dir)

    dataset = safe_load_datasets(
        dataset_name_or_path=dataset_name_or_path,
        split_name=dataset_split_name,
        cache_dir=cache_dir,
    )
    tokenizer = get_tokenizer(
        model_name_or_path=model_name_or_path, cache_dir=cache_dir
    )
    calib_dataloader = get_calib_dataloader(
        dataset=dataset,
        lookup_column_name=dataset_target_column_name,
        max_length=dataset_max_length,
        num_samples=dataset_num_samples,
        batch_size=dataset_batch_size,
        seed=seed,
        tokenizer=tokenizer,
    )

    quantize(
        model_name_or_path,
        mode,
        dry_run=dry_run,
        save_dir=output_dir,
        cache_dir=cache_dir,
        device=device,
        offload=offload,
        calib_dataloader=calib_dataloader,
        pedantic_level=pedantic_level,
    )

    msg = (
        f"Checkpoint({model_name_or_path}) can be converted."
        if dry_run
        else f"Checkpoint({model_name_or_path}) has been converted successfully."
    )
    typer.secho(msg)



# Copyright (c) 2024-present, FriendliAI Inc. All rights reserved.
# pylint: disable=no-value-for-parameter, no-name-in-module

"""FMO-CLI Dataset Utils."""
from __future__ import annotations

import os
from typing import Optional

import datasets  # type: ignore
import torch
from fmo_core import NotSupportedError, QuantizationError  # type: ignore
from torch.utils.data import DataLoader
from transformers import AutoTokenizer, PreTrainedTokenizer  # type: ignore


def safe_load_datasets(
    dataset_name_or_path: str,
    split_name: Optional[str],
    cache_dir: Optional[str] = None,
) -> datasets.Dataset:
    """Load dataset from calibration dataset config."""
    try:
        if os.path.exists(dataset_name_or_path):
            dataset = datasets.load_dataset(
                data_files=dataset_name_or_path,
                split=split_name,
            )
        else:
            data_name_parts = dataset_name_or_path.split(":")
            if len(data_name_parts) == 1:
                dataset = datasets.load_dataset(
                    dataset_name_or_path, split=split_name, cache_dir=cache_dir
                )
            elif len(data_name_parts) == 2:
                data_name, subset_name = data_name_parts
                dataset = datasets.load_dataset(
                    data_name, subset_name, split=split_name, cache_dir=cache_dir
                )
            else:
                raise QuantizationError(
                    "Dataset name is in invalid format. "
                    "(valid format: '<dataset_name>' or '<dataset_name>:<subset_name>')"
                )
    except ValueError as err:
        raise QuantizationError(f"datasets.load_dataset failed. {str(err)}") from err

    if not isinstance(dataset, datasets.Dataset):
        raise QuantizationError(
            "This dataset format is not supported for the calibration."
        )

    return dataset


def get_tokenizer(
    model_name_or_path: str,
    *,
    cache_dir: Optional[str] = None,
) -> PreTrainedTokenizer:
    """Try to get tokenizer of a pretrained model."""
    try:
        tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path,
            cache_dir=cache_dir,
            trust_remote_code=True,
        )
    except OSError as exc:
        raise NotSupportedError(str(exc)) from exc

    if not tokenizer.is_fast:
        raise NotSupportedError(
            "This model does not support Friendli-compatible tokenizer"
        )

    if tokenizer.pad_token != "<unk>":
        tokenizer.pad_token = tokenizer.eos_token
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token

    return tokenizer


def get_calib_dataloader(  # pylint: disable=too-many-arguments
    tokenizer: PreTrainedTokenizer,
    dataset: datasets.Dataset,
    lookup_column_name: str,
    seed: Optional[int] = 42,
    max_length: Optional[int] = 512,
    num_samples: Optional[int] = 512,
    batch_size: Optional[int] = 32,
) -> DataLoader:
    """Return Calibration DataLoader."""
    try:
        dataset = dataset.shuffle(seed=seed).select(range(num_samples * 2))  # type: ignore
        encoded_ds_w_special_tokens = tokenizer(
            dataset[lookup_column_name][:num_samples],
            return_tensors="pt",
            truncation=True,
            padding=True,
            max_length=max_length,
            add_special_tokens=True,
        ).input_ids
        encoded_ds_wo_special_tokens = tokenizer(
            dataset[lookup_column_name][num_samples:],
            return_tensors="pt",
            truncation=True,
            padding=True,
            max_length=max_length,
            add_special_tokens=False,
        ).input_ids

        max_length_diff = (
            encoded_ds_w_special_tokens.shape[-1]
            - encoded_ds_wo_special_tokens.shape[-1]
        )
        if max_length_diff > 0:
            padded_tokens = torch.full(
                (encoded_ds_wo_special_tokens.shape[0], max_length_diff),
                tokenizer.pad_token_id,
            )
            encoded_ds_wo_special_tokens = torch.cat(
                [encoded_ds_wo_special_tokens, padded_tokens], dim=1
            )
        assert (
            encoded_ds_w_special_tokens.shape[-1]
            == encoded_ds_wo_special_tokens.shape[-1]
        )
        encoded_dataset = torch.cat(
            [encoded_ds_w_special_tokens, encoded_ds_wo_special_tokens], dim=0
        )

    except KeyError as exc:
        raise NotSupportedError(
            f"`{lookup_column_name}` is not valid column name in given dataset."
        ) from exc

    return DataLoader(encoded_dataset, batch_size=batch_size)  # type: ignore
